from typing import Any, Union
import torch 
import torch.nn as nn
from torch.distributions import MultivariateNormal
import gpytorch as gpt

class GPXRegressorModule(nn.Module):
    """GPX Regression Module.

    Parameters
    ----------
    kernel : object
        A kernel function module implemented in gpytorch.kernels.
    kernel_kwargs : dict
        A dictionary of parameters for ``kernel``.
    dtype : object or string
        dtype.
    """
    def __init__(
        self, 
        kernel: Any = gpt.kernels.RBFKernel, 
        kernel_kwargs: dict = {}, 
        dtype: torch.dtype = torch.double
    ):
        super(GPXRegressorModule, self).__init__()
        self.dtype = dtype
        self.kernel = kernel
        self.kernel_kwargs = kernel_kwargs
        self.kernel_obj = gpt.kernels.ScaleKernel(kernel(**kernel_kwargs))
        self._init_noise_w = 0.1 
        self._init_noise_y = 0.1

    def train_initialize(
        self, X: torch.Tensor, Y: torch.Tensor, Z: torch.Tensor
    ):
        self.X_tr = X
        self.Y_tr = Y
        self.Z_tr = Z
        self.n_samples = X.size(0)
        self.x_dim = X.size(1)
        self.z_dim = Z.size(1)
        self._register_initialized_parameters(X, Y, Z)
        self._ZZ = torch.matmul(Z, torch.t(Z))  # (n, n)

    def _register_initialized_parameters(
        self, X: torch.Tensor, Y: torch.Tensor, Z: torch.Tensor
    ):
        noise_w = nn.Parameter(torch.tensor(self._init_noise_w, dtype=self.dtype))
        noise_y = nn.Parameter(torch.tensor(self._init_noise_y, dtype=self.dtype))
        self.register_parameter("noise_w", noise_w)
        self.register_parameter("noise_y", noise_y)

    def forward(
        self, X: torch.Tensor, Y: torch.Tensor, Z: torch.Tensor
    ):
        """Forward calculation (calculating negative log Gaussian likelihood).

        Parameters
        ----------
        X : torch.tensor
            training data of original representation with size of (n_samples, n_x_features).
        Y : torch.tensor
            training data of targets with size of (n_samples, 1).
        Z : torch.tensor
            training data of simplified representation with size of (n_samples, n_z_features).
        """
        ZZ = self._ZZ if hasattr(self, "_ZZ") else None
        loss = self.neg_log_likelihood(X, Y, Z, ZZ) / X.shape[0]
        return loss

    def prepare_eval(self):
        C = self._compute_covar_y(self.X_tr, self.Z_tr, self._ZZ)
        self.C_inv = torch.cholesky_inverse(torch.linalg.cholesky(C))
        K = self._compute_kernel_matrix(self.X_tr, self.Z_tr)  # (n, n)
        self.KIinv = torch.inverse(K + self.noise_w**2 * torch.eye(self.n_samples))  # (n, n)
        self.ZKZ = (K + self.noise_w**2 * torch.eye(self.n_samples, dtype=self.dtype)) * self._ZZ # (n, n)

    def _compute_kernel_matrix(
        self, X: torch.Tensor, Z: torch.Tensor, 
        X2: Union[torch.Tensor, None] = None, Z2: Union[torch.Tensor, None] = None
    ):
        S = self.kernel_obj(X, X2).evaluate().type(self.dtype)  # (n, n')
        return S

    def neg_log_likelihood(
        self, X: torch.Tensor, Y: torch.Tensor, Z: torch.Tensor, 
        ZZ: Union[torch.Tensor, None] = None
    ):
        """Calculating negative log Gaussian likelihood.
        """
        if ZZ == None:
            ZZ = torch.matmul(Z, torch.t(Z))  # (n, n)
        covar_y = self._compute_covar_y(X, Z, ZZ)
        b = torch.zeros(self.n_samples, dtype=self.dtype)
        mn = MultivariateNormal(b, covar_y)
        nll = -mn.log_prob(Y)
        return nll

    def _compute_covar_y(
        self, X: torch.Tensor, Z: torch.Tensor, ZZ: torch.Tensor
    ):
        n = Z.size(0)
        K = self._compute_kernel_matrix(X, Z)
        covar = self.noise_y**2 * torch.eye(n, dtype=self.dtype)
        covar += K * ZZ
        covar += self.noise_w**2 * torch.diag(ZZ.diagonal())
        return covar  # (n, n)

    def predict_targets(
        self, X_new: torch.Tensor, Z_new: torch.Tensor
    ):
        """Predicting target variables.
        """
        X_all = torch.cat((self.X_tr, X_new), 0)
        Z_all = torch.cat((self.Z_tr, Z_new), 0)
        ZZ_all = torch.matmul(Z_all, torch.t(Z_all))  # (n_all, n_all)
        C_all = self._compute_covar_y(X_all, Z_all, ZZ_all)
        k = C_all[self.n_samples:][:, :self.n_samples]  # (n_new, n)
        Ey = torch.matmul(torch.matmul(k, self.C_inv), self.Y_tr)
        C_remain = C_all[self.n_samples:][:, self.n_samples:]  # (n_new, n_new)
        Vy = C_remain - torch.matmul(torch.matmul(k, self.C_inv), torch.t(k))
        return Ey, Vy

    def predict_weights(
        self, X_new: torch.Tensor, Z_new: torch.Tensor
    ):
        """Estimating local weights.
        """
        # inverse = lambda x: torch.cholesky_inverse(torch.cholesky(x))
        # n_new = Z_new.size(0)
        X_all = torch.cat((self.X_tr, X_new), 0)
        Z_all = torch.cat((self.Z_tr, Z_new), 0)
        K_all = self._compute_kernel_matrix(X_all, Z_all)  # (n+n_new, n+n_new)
        K = K_all[:self.n_samples][:, :self.n_samples]  # (n, n)
        swI = self.noise_w**2 * torch.eye(self.n_samples, dtype=self.dtype)  # (n, n)
        K += swI
        k = K_all[self.n_samples:][:, :self.n_samples]  # (n_new, n)
        A = torch.matmul(k, self.KIinv)  # (n_new, n)
        KZ = torch.matmul(K.expand(self.z_dim, -1, -1), torch.diag_embed(torch.t(self.Z_tr)))  # (z_dim, n, n)
        AKZ = torch.bmm(A.expand(self.z_dim, -1, -1), KZ)  # (z_dim, n_new, n)
        # sy2I = self.noise_y**2 * torch.eye(self.n_samples, dtype=self.dtype)
        AKZDinvZKZ = torch.matmul(torch.matmul(AKZ, self.C_inv), self.ZKZ)  # (z_dim, n_new, n)

        M = self.noise_y**(-2) * torch.matmul((AKZ - AKZDinvZKZ), self.Y_tr)  # (z_dim, n_new)
        M = torch.t(M)  # (n_new, z_dim)

        c = torch.diagonal(K_all[self.n_samples:][:, self.n_samples:]) + self.noise_w**2 # (n_new,)
        c = torch.diag_embed(torch.t(c.expand(self.z_dim, -1)))  # (n_new, z_dim, z_dim)
        Ak = torch.matmul(A.unsqueeze(-2), k.unsqueeze(-1)).flatten()  # (n_new)
        Ak = torch.diag_embed(torch.t(Ak.expand(self.z_dim, -1)))  # (n_new, z_dim, z_dim)
        AK = torch.matmul(A, K)  # (n_new, n) 
        AKA = torch.matmul(AK.unsqueeze(-2), A.unsqueeze(-1)).flatten()  # (n_new)
        AKA = torch.diag_embed(torch.t(AKA.expand(self.z_dim, -1)))  # (n_new, z_dim, z_dim)
        AKZDinvZKA = torch.matmul(
            torch.matmul(AKZ.permute(1, 0, 2), self.C_inv), AKZ.permute(1, 2, 0))  # (n_new, z_dim, z_dim)
        V = c - Ak + AKA - AKZDinvZKA

        return M, V


import numpy as np 
import torch

def to_tensor(*args):
    """
    Convert args to tensors
    """
    def convert(var):
        if isinstance(var, np.ndarray):
            return torch.from_numpy(var)
        elif isinstance(var, torch.Tensor):
            return var
        else:
            raise NotImplementedError("Invalid data structure or type")

    converted_vars = map(lambda x: convert(x), args)
    return converted_vars


def to_ndarray(*args):
    """
    Convert args to ndarray
    """
    def convert(var):
        if isinstance(var, torch.Tensor):
            return var.to("cpu").detach().numpy().copy()
        elif isinstance(var, np.ndarray):
            return var
        else:
            raise NotImplementedError("Invalid data structure or type")

    converted_vars = map(lambda x: convert(x), args)
    return converted_vars



from typing import Union
import numpy as np
import torch

def calc_rbf_bandwidth(X: Union[np.ndarray, torch.Tensor]):
    if not isinstance(X, np.ndarray):
        X = X.cpu().numpy()
    n = X.shape[0]
    G = np.sum(X * X, 1).reshape(n, 1)
    Q = np.tile(G, (1, n))
    R = np.tile(G.T, (n, 1))

    dists = Q + R - 2 * np.dot(X, X.T)
    dists = dists - np.tril(dists)
    dists = dists.reshape(n**2, 1)

    bandwidth = np.sqrt(0.5 * np.median(dists[dists > 0]))
    return bandwidth
    

def calc_rbf_gamma(X_tensor: torch.Tensor):
    bandwidth = calc_rbf_bandwidth(X_tensor)
    gamma = 0.5 * bandwidth**-2
    return gamma
    


from typing import Any, Union
import numpy as np
import torch
import gpytorch as gpt
from sklearn.base import RegressorMixin, BaseEstimator
from sklearn.metrics import r2_score

from .gpx_regressor_module import GPXRegressorModule
from .tensor_utils import to_ndarray, to_tensor


class GPXRegressor(RegressorMixin, BaseEstimator):
    """GPX for regression. 
    """
    def __init__(
        self, 
        kernel: Any = gpt.kernels.RBFKernel, 
        kernel_kwargs: dict = {}, 
        kernel_init_params: dict = {},
        max_iter: int = 150, 
        tol: float = 10**-4, 
        lr: float = 0.1,
        dtype: torch.dtype = torch.double, 
        verbose: bool = False,
    ):
        self.kernel = kernel
        self.kernel_kwargs = kernel_kwargs
        self.kernel_init_params = kernel_init_params
        self.max_iter = max_iter
        self.tol = tol 
        self.lr = lr
        self.dtype = dtype
        self.verbose = verbose 

        self.model = GPXRegressorModule(kernel, kernel_kwargs=kernel_kwargs, dtype=dtype)
        for k, v in kernel_init_params.items():
            self.model.kernel_obj.base_kernel.__dict__[k] = v


    def fit(self, X: np.ndarray, y: np.ndarray, Z: Union[np.ndarray, None]):
        """Train model.

        Parameters
        ----------
        X : numpy.ndarray
        y : numpy.ndarray
        Z : numpy.ndarray or None
        """
        Z = X if Z is None else Z
        X, y, Z = map(lambda x: x.type(self.dtype), to_tensor(X, y, Z))

        self.model.type(self.dtype)
        self.model.train_initialize(X, y, Z)
        self.model.train()
        if self.verbose:
            for param_name, param in self.model.named_parameters():
                print(f'Parameter name: {param_name:42} value = {param.item()}')

        # use LBFGS as optimizer since we can load the whole data to train
        def closure():
            optimizer.zero_grad()
            loss = self.model(X, y, Z)
            if self.verbose:
                print('Iter {0:3d}: loss ='.format(optimizer.iter_count), loss.item())
            optimizer.iter_count += 1
            loss.backward()
            return loss

        optimizer = torch.optim.LBFGS(
            self.model.parameters(), lr=self.lr, tolerance_change=self.tol, max_iter=self.max_iter)
        optimizer.iter_count = 1
        optimizer.step(closure)

        if self.verbose:
            for param_name, param in self.model.named_parameters():
                print(f'Parameter name: {param_name:42} value = {param.item()}')

        self.model.prepare_eval()
        return self


    def predict(
        self, X: np.ndarray, Z: Union[np.ndarray, None] = None, return_weights: bool = False
    ):
        """Prediction.

        Parameters
        ----------
        X : numpy.ndarray
        Z : numpy.ndarray or None
        return_weights : bool
            Decide whether to return sample-wise weights. 
        """

        Z = X if Z is None else Z
        X, Z = map(lambda x: x.type(self.dtype), to_tensor(X, Z))
        y_mean, y_cov = to_ndarray(*self.model.predict_targets(X, Z))
        if return_weights:
            w_mean, w_cov = to_ndarray(*self.model.predict_weights(X, Z))
            return y_mean, y_cov, w_mean, w_cov
        else:
            return y_mean, y_cov


    def score(
        self, X: np.ndarray, y: np.ndarray, Z: Union[np.ndarray, None] = None, 
        sample_weight: Union[np.ndarray, None] = None
    ):
        y_pred, _ = self.predict(X, Z)
        return r2_score(y, y_pred, sample_weight=sample_weight)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals
import os

from setuptools import setup, find_packages

try:
    with open('README.md') as f:
        readme = f.read()
except IOError:
    readme = ''

def _requires_from_file(filename):
    return open(filename).read().splitlines()

# version
here = os.path.dirname(os.path.abspath(__file__))
version = '0.1'

setup(
    name="pytorch-gpx",
    version=version,
    url='https://github.com/yuyay/gpx',
    author='Yuya Yoshikawa',
    author_email='yoshikawa@stair.center',
    maintainer='Yuya Yoshikawa',
    maintainer_email='yoshikawa@stair.center',
    description='An official implementation of "Gaussian Process Regression With Interpretable Sample-Wise Feature Weights"',
    # long_description=readme,  #TODO: convert to reST format
    packages=find_packages(
        exclude=('examples', 'docs', '.vscode', 'docker', 'scripts', 'experiments', 'data', 'notebooks')),
    python_requires='>=3.7',
    install_requires=_requires_from_file('requirements.txt'),
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # entry_points="""
    #   # -*- Entry points: -*-
    #   [console_scripts]
    #   pkgdep = pypipkg.scripts.command:main
    # """,
)

# from setuptools import setup, find_packages

# setup(
#     name="gpx",
#     version="0.9",
#     install_requires=[
#         "numpy", "scipy", "torch", "matplotlib", "scikit-learn",
#         "jupyter", "ipython", "gpytorch"],
#     packages=find_packages(
#         exclude=('examples', 'docs', '.vscode', 'docker', 'scripts', 'experiments', 'data')),
# )
